import * as browser from "webextension-polyfill";
import * as ewe from '@eyeo/webext-ad-filtering-solution';
import { parseUri } from "./utils/utils.js";
import {
  storageLocalGet,
  storageLocalRemove,
  storageLocalSet,
  isDataCorrupted,
} from "./storage.js";
import { isFirstRun } from "./first-run.js";
import { openInstalled } from "./slash-installed.js";

// storage key paused state is saved under
export const PAUSED_KEY = "ub_paused";
// allow-list all blocking requests regardless of frame / document, but still allows element hiding
const PAUSED_FILTER_TEXT1 = "@@*";
// allow-list all documents, which prevents element hiding
const PAUSED_FILTER_TEXT2 = "@@*$document";

/**
 * The ad blocking manager.
 */
export class Blocker {
  constructor(name, version) {
    this.name = name;
    this.version = version;
  }

  /**
   * Checks if the given URL can have the ad blocker running on it.
   *
   * @param url {string|URL} the url to check
   * @returns {boolean} true if the ad blocker cannot be blocked
   */
  isPageUnblockable(url) {
    if (!url) {
      // Protect against empty URLs - e.g. Safari empty/bookmarks/top sites page
      return true;
    }
    let scheme;
    if (!url.protocol) {
      scheme = parseUri(url).protocol;
    } else {
      scheme = url.protocol;
    }
    return scheme !== "http:" && scheme !== "https:" && scheme !== "feed:";
  }

  /**
   * Get or set if uBlock is paused
   *
   * @param [newValue] {boolean} undefined to retrieve paused state, true if uBlock will be paused, false to unpause.
   * @returns {boolean|undefined} undefined if newValue was specified, otherwise, it returns true if paused, false otherwise
   */
  async isPaused(newValue) {
    if (newValue === undefined) {
      const paused = await storageLocalGet(PAUSED_KEY);
      return paused[PAUSED_KEY] === true;
    }
    if (newValue === true) {
      await storageLocalSet({ [PAUSED_KEY]: true });
      await ewe.filters.add([PAUSED_FILTER_TEXT1], {});
      await ewe.filters.add([PAUSED_FILTER_TEXT2], {});
    } else {
      await ewe.filters.remove([PAUSED_FILTER_TEXT1]);
      await ewe.filters.remove([PAUSED_FILTER_TEXT2]);
      await storageLocalRemove(PAUSED_KEY);
    }
    return undefined;
  }

  /**
   * Creates a custom filter entry that allow-lists a given domain
   * @param pageUrl {string} url of the page
   * @param origin {string} the source or trigger for the filter list entry
   * @return {Promise<*|null|undefined>} null if successful, otherwise an exception
   */
  async createDomainAllowListFilter(pageUrl, origin) {
    const theURL = new URL(pageUrl);
    const host = theURL.hostname.replace(/^www\./, "");
    const filter = `@@||${host}/*^$document`;
    return this.addCustomFilter(filter, origin);
  }

  /**
   * Look for a custom filter that would allow-list the 'url' parameter
   * and if any exist, remove the first one.
   * @param pageUrl {string} a URL that may be allow-listed by a custom filter
   * @param tabId {int} tab id of the tab that may be allow-listed by a custom filter
   * @return {Promise<boolean>} true if a filter was found and removed; false otherwise.
   */
  async tryToRemoveAllowList(pageUrl, tabId) {
    const url = pageUrl.replace(/#.*$/, ""); // Whitelist ignores anchors
    const customFilters = await this.getUserFilters();
    if (!customFilters || customFilters.length === 0) {
      return false;
    }

    /* eslint-disable no-await-in-loop */
    for (let i = 0; i < customFilters.length; i++) {
      const { text } = customFilters[i];
      const allow = text.search(/@@\*\$document,domain=~/);
      // Remove allow-list site, which is allow-listed by global @@*&document,domain=~
      // filter
      if (allow > -1) {
        // Remove protocols
        const [finalUrl] = url
          .replace(/((http|https):\/\/)?(www.)?/, "")
          .split(/[/?#]/);
        const mainFilters = [text];
        const secondaryFilters = [`${text}|~${finalUrl}`];
        await ewe.filters.remove(mainFilters);
        await ewe.filters.remove(secondaryFilters);
        return true;
      }
      if (
        this.isAllowlistFilter(text) &&
        (await ewe.filters.getAllowingFilters(tabId)).includes(text)
      ) {
        const mainFilters = [text];
        await ewe.filters.remove(mainFilters);
        return true;
      }
    }
    return false;
  }

  /**
   * Checks if the given filter is an allow-list filter rule.
   * @param text {string} the filter rule to check
   * @return {boolean} true if the rule is an allow-list rule.
   */
  isAllowlistFilter(text) {
    return /^@@/.test(text);
  }

  /**
   * Retrieve all user filters.
   *
   * @return {Promise<*>} promise resolving to all the user filters.
   */
  async getUserFilters() {
    return ewe.filters.getUserFilters();
  }

  /**
   * Add a new custom filter entry.
   * @param filterText {string} line of text to add to custom filters.
   * @param origin {string} the source or trigger for the filter list entry
   * @return {Promise<string|null|*>} null if successful, otherwise an exception
   */
  async addCustomFilter(filterText, origin) {
    try {
      const response = ewe.filters.validate(filterText);
      if (response) {
        return response;
      }
      await ewe.filters.add([filterText], this.createFilterMetaData(origin));
      await ewe.filters.enable([filterText]);
      return null;
    } catch (ex) {
      // convert to a string so that Safari can pass
      // it back to content scripts
      return ex.toString();
    }
  }

  /**
   * Creates the meta-data to be saved with a users custom filter rules.
   *
   * - created - an integer representing the number of milliseconds elapsed since January 1, 1970 00:00:00 UTC.
   * - origin - a string representing the method the user added the filter rule
   *
   * @param [origin] {string} a string representing the method the user added the filter rule
   * @return {{created:number}|{created:number, origin:string}} a new meta-data object for the filter containing a created timestamp
   */
  createFilterMetaData(origin) {
    const data = { created: Date.now() };
    if (origin) {
      data.origin = origin;
    }
    return data;
  }

  /**
   * Add subscriptions on first run and configure AA.
   */
  async addSubscriptions() {
    if (isFirstRun()) {
      // // If we need to do something on first run, do it here
      try {
        await ewe.subscriptions.addDefaults();
      }
      catch (ex) {
        console.error("Failed to add default filter lists:", ex);

        // We don't want to keep the extension in a broken state, so we just add EasyList
        ewe.subscriptions
          .add("https://easylist-downloads.adblockplus.org/v3/full/easylist.txt")
          .then(() => {
            // do we care if it's installed
          })
          .catch((err) => {
            console.log("error adding lists", err);
            // do we do something with this error?
          });
      }
      if (await ewe.subscriptions.has("https://easylist-downloads.adblockplus.org/v3/full/exceptionrules.txt")) {
        await ewe.subscriptions.remove("https://easylist-downloads.adblockplus.org/v3/full/exceptionrules.txt");
      }
      if (await ewe.subscriptions.has("https://easylist-downloads.adblockplus.org/exceptionrules.txt")) {
        await ewe.subscriptions.remove("https://easylist-downloads.adblockplus.org/exceptionrules.txt");
      }
    }

    // add the Anti-CV filter list for all users (existing)
    ewe.subscriptions
      .add("https://easylist-downloads.adblockplus.org/v3/full/abp-filters-anti-cv.txt")
      .then(() => {
        // do we care if it's installed
      })
      .catch((err) => {
        console.log("error adding lists", err);
        // do we do something with this error?
      });

    // Show first run page or the updates page. The latter is only shown
    // on Chromium (since the current updates page announces features that
    // aren't new to Firefox users), and only if this version of the
    // updates page hasn't been shown yet.

    // Always show the first run page if a data corruption was detected
    // (either through failure of reading from or writing to storage.local).
    // The first run page could notify the user about the data corruption.
    if (isFirstRun() || isDataCorrupted()) {
      // TODO we may also want to clean out any old storage data that we aren't using
      openInstalled();
    }
  }

  /**
   * Remove any subscriptions that a user or administrator has added to a
   * central / common configuration (such as the Windows Registry)
   *
   * @return {Promise}
   */
  removeSubscriptions() {
    return new Promise((resolve) => {
      if ("managed" in browser.storage) {
        storageLocalGet(null).then(
          async (items) => {
            for (let key in items) {
              if (
                key === "remove_subscriptions" &&
                Array.isArray(items[key]) &&
                items[key].length
              ) {
                for (let inx = 0; inx < items[key].length; inx++) {
                  await ewe.subscriptions.remove(items[key][inx]);
                }
              }
            }
            resolve();
          },

          // Opera doesn't support browser.storage.managed, but instead of simply
          // removing the API, it gives an asynchronous error which we ignore here.
          () => {
            resolve();
          }
        );
      } else {
        resolve();
      }
    });
  }

  /**
   * Obtain meta-data about the current tab and the extension's state on that tab.
   *
   * info object passed to resolve: {
   * page: Page object
   * tab: Tab object
   * allowListed: bool - whether the current tab's URL is allow-listed.
   * disabled_site: bool - true if the url is e.g. about:blank or the
   * Extension Gallery, where extensions don't run.
   * settings: Settings object
   * paused: bool - whether AdBlock is paused
   * domainPaused: bool - whether the current tab's URL is paused
   * blockCountPage: int - number of ads blocked on the current page
   * blockCountTotal: int - total number of ads blocked since install
   * customFilterCount: int - number of custom rules for the current tab's URL
   * showMABEnrollment: bool - whether to show MAB enrollment
   * popupMenuThemeCTA: string - name of current popup menu CTA theme
   * lastGetStatusCode: int - status code for last GET request
   * lastGetErrorResponse: error object - error response for last GET request
   * lastPostStatusCode: int - status code for last POST request
   * allowListRuleText: string - allow-list rule text for use on YouTube and Twitch
   * }
   *
   * @param secondTime {boolean} true if the method is being called recursively
   * @param [tabId] {int} the tab to obtain data about, or undefined for the current active tab
   * @returns {Promise<{stack: *, message, errorStr: string}|{}|{disabledSite: boolean, settings: *, paused: (boolean|undefined), domainPaused: (any|boolean), id, url: string}>}
   */
  async getTabInfo(secondTime, tabId) {
    const tab = await this.getTab(tabId);
    if (tab && !tab.url) {
      // Issue 6877: tab URL is not set directly after you opened a window
      // using window.open()
      if (!secondTime) {
        setTimeout(() => {
          return this.getTabInfo(true);
        }, 250);
      }
      return {};
    }
    try {
      const disabledSite = this.isPageUnblockable(tab.url);

      const result = {
        disabledSite,
        url: String(tab.url),
        id: tab.id,
        paused: await this.isPaused(),
      };
      if (!disabledSite) {
        result.allowListed = !!(await ewe.filters.getAllowingFilters(tab.id))
          .length;
        result.allowListedText = await ewe.filters.getAllowingFilters(tab.id);
      }
      return result;
    } catch (err) {
      return {
        errorStr: err.toString(),
        stack: err.stack,
        message: err.message,
      };
    }
  }

  /**
   * Get a tab by ID or the current focused tab if no ID provided.
   *
   * @param [tabId] {int} the ID of the tab to locate or undefined to retrieve the current active tab
   * @returns {Promise<object>} the tab found (https://developer.mozilla.org/en-US/docs/Mozilla/Add-ons/WebExtensions/API/tabs/Tab)
   */
  async getTab(tabId) {
    if (tabId) {
      let id = tabId;
      if (typeof id === "string") {
        id = parseInt(id, 10);
      }
      return await browser.tabs.get(id);
    } else {
      const tabs = await browser.tabs.query({
        active: true,
        lastFocusedWindow: true,
      });
      if (tabs.length === 0) {
        return; // For example: only the background devtools or a popup are opened
      }
      return tabs[0];
    }
  }

  // ---------------------------------------------------------------------
  // QA Support Features
  // ---------------------------------------------------------------------

  async subscriptionsGet({ ignoreDisabled, disabledFilters }) {
    let subscriptions = [];
    for (let s of await ewe.subscriptions.getDownloadable()) {
      if (ignoreDisabled && !s.enabled) continue;

      let subscription = toPlainSubscription(s);
      if (disabledFilters) {
        let filters = (await ewe.subscriptions.getFilters(s.url)) || [];
        subscription.disabledFilters = filters
          .filter((f) => f.enabled === false)
          .map((f) => f.text);
      }
      subscriptions.push(subscription);
    }
    return subscriptions;
  }

  async subscriptionsRemove({ url }) {
    return ewe.subscriptions.remove(url);
  }

  async filtersGet() {
    let filters = await ewe.filters.getUserFilters();
    return filters.map(toPlainFilter);
  }

  async filtersRemove({ text }) {
    await ewe.filters.remove([text]);
    // in order to behave, from consumer perspective, like any other
    // method that could produce errors, return an Array, even if empty
    return [];
  }

  async filtersImportRaw({ text, origin }) {
    let [filterTexts, errors] = this.filtersValidate(text);

    if (errors.length === 0) {
      for (const filterText of filterTexts) {
        const error = await this.addFilter(filterText, origin);
        if (error) errors.push(error);
      }
    }
    return errors;
  }

  filtersValidate(text) {
    let filterTexts = [];
    let errors = [];

    let lines = text.split("\n");
    for (let i = 0; i < lines.length; i++) {
      let [filterText, error] = this.parseFilter(lines[i]);

      if (error) {
        // We don't treat filter headers like invalid filters,
        // instead we simply ignore them and don't show any errors
        // in order to allow pasting complete filter lists.
        // If there are no filters, we do treat it as an invalid filter
        // to inform users about it and to give them a chance to edit it.
        if (error.type === "unexpected_filter_list_header" && lines.length > 1)
          continue;

        if (lines.length > 1) error.lineno = i + 1;

        errors.push(error);
      } else if (filterText) {
        filterTexts.push(filterText);
      }
    }

    return [filterTexts, errors];
  }

  parseFilter(text) {
    let filterText = text.trim() || null;
    let error = null;

    if (filterText) {
      if (filterText[0] === "[") {
        error = { type: "unexpected_filter_list_header" };
      } else {
        let filterError = ewe.filters.validate(filterText);
        if (filterError) error = toPlainFilterError(filterError);
      }
    }

    return [filterText, error];
  }

  async addFilter(filterText, origin) {
    const data = { created: Date.now() };
    if (origin) data.origin = origin;

    try {
      await ewe.filters.add([filterText], data);
      await ewe.filters.enable([filterText]);
    } catch (ex) {
      // If we add an existing filter again, we ignore the resulting error to
      // avoid showing unnecessary error messages to the user
      if (ex.type !== "storage_duplicate_filters") return ex;
    }

    return null;
  }
}

function toPlainObject(keys, obj) {
  let result = {};
  for (let key of keys) {
    if (key in obj) result[key] = obj[key];
  }
  return result;
}

function toPlainFilter(filter) {
  let obj = toPlainObject(["enabled", "slow", "text", "type"], filter);

  // For the time being, we are renaming the enabled property to
  // make the UI compatible with EWE without having to rename it
  // in the UI code itself just yet
  // For the same reason, we're not adding the property for comment filters
  if (obj.enabled !== null) obj.disabled = !obj.enabled;
  delete obj.enabled;

  return obj;
}

function toPlainSubscription(subscription) {
  let obj = toPlainObject(
    [
      "downloadable",
      "downloadStatus",
      "enabled",
      "homepage",
      "version",
      "lastDownload",
      "lastSuccess",
      "softExpiration",
      "expires",
      "title",
      "url",
      "downloading",
    ],
    subscription
  );

  // For the time being, we are renaming the enabled property to
  // make the UI compatible with EWE without having to rename it
  // in the UI code itself just yet
  obj.disabled = !obj.enabled;
  delete obj.enabled;

  return obj;
}

export let toPlainFilterError = toPlainObject.bind(null, [
  "lineno",
  "option",
  "reason",
  "type",
]);
