import * as browser from "webextension-polyfill";
import * as ewe from '@eyeo/webext-ad-filtering-solution';

import { storageLocalGet, storageLocalSet } from "./storage.js";

const SUBS_KEY = "ub_subs";
const SUBS_ALARM = "schedule.sync";

const instances = [];

const listener = async (alarm) => {
  if (alarm.name === SUBS_ALARM) {
    // console.log("sync alarm matched", SUBS_ALARM, alarm);
    instances.forEach((s) => {
      // console.log("sync alarm handled");
      s.sync();
    });
  } else {
    // console.log("sync alarm skipped", alarm);
  }
};

export class Subscriptions {
  /**
   * Create a new Subscription without options.
   */
  constructor() {
    instances.push(this);
  }

  /**
   * Sync begins an asynchronous download of filter subscriptions and schedules the next sync event.
   *
   * @return {Promise<void>} a promise that resolves when the subscription sync begins and another sync event is scheduled.
   */
  async sync() {
    const settings = await this.getSubSettings();
    settings.totalSyncs = settings.totalSyncs + 1;
    await storageLocalSet({ [SUBS_KEY]: settings });

    await ewe.subscriptions.sync();
  }

  /**
   * Obtain the subscription settings.
   *
   * @return {Promise<*|{subscriptions: {}, totalSyncs: number}>} a promise that resolves with the subscription settings for the extension.
   */
  async getSubSettings() {
    const data = await storageLocalGet(SUBS_KEY);
    return (
      (data && data[SUBS_KEY]) || {
        subscriptions: {},
        totalSyncs: 0,
      }
    );
  }

  /**
   * Schedule the next sync event - should be called once during initialization to
   * schedule the first sync event. Once sync has occurred, the system calls itself
   * forever.
   *
   * @return {Promise<void>} resolves when the sync event has been scheduled.
   */
  async scheduleSyncEvent() {
    if (browser.alarms.onAlarm.hasListener(listener)) {
      // console.log("sync listener exists");
    } else {
      // console.log("sync listener added");
      browser.alarms.onAlarm.addListener(listener);
    }

    const existing = await browser.alarms.get(SUBS_ALARM);
    if (existing && existing.name === SUBS_ALARM) {
      // already scheduled
      // console.log("sync alarm exists", SUBS_ALARM, existing);
      return;
    }
    browser.alarms.create(SUBS_ALARM, {
      delayInMinutes: 7,
      periodInMinutes: 60 * 24,
    });
    // console.log("sync alarm created", SUBS_ALARM);
  }
}
